import re
from contextlib import contextmanager
from typing import MutableMapping, Tuple  # noqa
from pyparsing import (  # type: ignore
    alphas,
    alphanums,
    nums,
    restOfLine,
    srange,
    CaselessKeyword,
    Forward,
    Group,
    Keyword,
    Literal,
    Optional,
    ParseException,
    ParserElement,
    QuotedString,
    Word,
    ZeroOrMore,
)
from . import asp
from . import input as i
from . import output as o

__all__ = [
    'parse_input_spec',
    'parse_output_spec',
    'parse_spec',
    'parse_embedded_spec',
    'parse_answer_set',
]


@contextmanager
def PyParsingDefaultWhitespaceChars(whitespace_chars):
    '''Set the given whitespace_chars as pyparsing's default whitespace chars while the context manager is active.

    Since ParserElement.DEFAULT_WHITE_CHARS is a global variable, this method is not thread-safe (but no pyparsing parser construction is thread-safe for the same reason anyway).
    '''
    # A possible solution to this problem:
    # Since the pyparsing code is basically a single big file, we could just copy it (under aspio/vendor or something like that) and have our own "private" version of pyparsing. (TODO: think about this some more and maybe do it)
    previous_whitespace_chars = ParserElement.DEFAULT_WHITE_CHARS
    ParserElement.setDefaultWhitespaceChars(whitespace_chars)
    yield
    ParserElement.setDefaultWhitespaceChars(previous_whitespace_chars)


DEFAULT_WHITESPACE_CHARS = ' \n\t\r'  # this is the same as pyparsing's default


def ignore_comments(parser):
    '''Ignore comments (starting with '%' and continuing until the end of the same line) on the given parser (ParserElement instance).'''
    comment = '%' + restOfLine
    parser.ignore(comment)
    return parser


# Common syntax elements
with PyParsingDefaultWhitespaceChars(DEFAULT_WHITESPACE_CHARS):
    alphas_lowercase = srange('[a-z]')
    alphas_uppercase = srange('[A-Z]')
    predicate_name = (Optional('-') + Word(alphas_lowercase, alphanums + '_')).setParseAction(''.join).setName('predicate name')
    # Currently we only support ASCII identifiers for the python side.
    # Python (starting with version 3.0) supports additional characters in identifiers, see https://docs.python.org/3/reference/lexical_analysis.html#identifiers
    # It would be nice to support the same set, but it's not absolutely necessary.
    py_identifier = Word(alphas + '_', alphanums + '_').setName('python identifier')
    py_qualified_identifier = Word(alphas + '_', alphanums + '_.').setName('qualified python identifier')
    integer = (Optional('-') + Word(nums)).setName('integer').setParseAction(lambda t: int(t[0]))
    positive_integer = Word(nums).setName('integer').setParseAction(lambda t: int(t[0]))
    lpar = Literal('(').suppress()
    rpar = Literal(')').suppress()
    lbracket = Literal('[').suppress()
    rbracket = Literal(']').suppress()
    lbrace = Literal('{').suppress()
    rbrace = Literal('}').suppress()
    langle = Literal('<').suppress()
    rangle = Literal('>').suppress()
    dot = Literal('.').suppress()
    comma = Literal(',').suppress()
    colon = Literal(':').suppress()
    semicolon = Literal(';').suppress()
    equals = Literal('=').suppress()
    amp = Literal('&').suppress()
    slash = Literal('/').suppress()
    rightarrow = Literal('->').suppress()

# TODO
# Improve error messages of all parsers!
# See http://blog.ezyang.com/2014/05/parsec-try-a-or-b-considered-harmful/ and check how much of that applies here.
# Also: http://stackoverflow.com/questions/33708817/parser-errors-pattern-for-generating-error-handling-automatically (-> "PEG" parser?)


def RawInputSpecParser():
    '''Syntax of the INPUT statement (and nothing else).'''
    with PyParsingDefaultWhitespaceChars(DEFAULT_WHITESPACE_CHARS):
        INPUT = CaselessKeyword('INPUT').suppress()
        FOR = CaselessKeyword('for').suppress()
        IN = CaselessKeyword('in').suppress()

        target = Forward()
        anonymous_var = Keyword('_')
        # Keywords cannot be used as variable names (we still allow "INPUT" as it never occurs inside the spec)
        input_keyword = FOR | IN
        var = (~input_keyword + ~anonymous_var + Word(alphas + '_', alphanums + '_')).setName('variable')
        tuple_match = lpar + target + ZeroOrMore(comma + target) + Optional(comma) + rpar
        # The target of an assignment, supporting tuple unpacking as a simple form of pattern matching in addition to plain variables
        target << (anonymous_var | var | tuple_match)
        #
        anonymous_var.setParseAction(lambda: i.AnonymousVariable())
        var.setParseAction(lambda t: i.Variable(str(t[0])))
        tuple_match.setParseAction(lambda t: i.TupleMatch(t))

        # Accessing objects, some examples:
        # - just access a variable directly:            node
        # - access a field on a variable:               node.label
        # - accessing a fixed index in a collection:    some_tuple[3]
        # - chainable:                                  node.neighbors[2].label
        field_accessor = dot + py_identifier('name')
        subscript = integer | QuotedString('"', escChar='\\')
        subscript_accessor = lbracket + subscript('key') + rbracket
        accessor = var('var') + Group(ZeroOrMore(field_accessor | subscript_accessor))('path')
        #
        field_accessor.setParseAction(lambda t: i.Attribute(t.name))
        subscript_accessor.setParseAction(lambda t: i.Subscript(t.key))
        accessor.setParseAction(lambda t: i.Accessor(t.var, t.path))

        # Iterating over objects
        iteration = FOR + target('target') + IN + accessor('accessor')
        iterations = Group(ZeroOrMore(iteration))
        #
        iteration.setParseAction(lambda t: i.Iteration(t.target, t.accessor))

        predicate_args = Group(Optional(accessor + ZeroOrMore(comma + accessor) + Optional(comma)))
        predicate_spec = predicate_name('pred') + lpar + predicate_args('args') + rpar + iterations('iters') + semicolon
        predicate_specs = Group(ZeroOrMore(predicate_spec))
        #
        predicate_spec.setParseAction(lambda t: i.Predicate(t.pred, t.args, t.iters))

        # Allow optional types, e.g., Set<Node> etc.
        input_type = Forward()
        input_type << (py_qualified_identifier('type_name') + Group(Optional(langle + input_type + ZeroOrMore(comma + input_type) + rangle))('type_args'))
        input_arg = Group((input_type('type') + var('name')) | var('name'))
        input_args = Group(Optional(input_arg + ZeroOrMore(comma + input_arg) + Optional(comma)))

        input_statement = INPUT + lpar + input_args('args') + rpar + lbrace + predicate_specs('preds') + rbrace
        #
        input_statement.setParseAction(lambda t: i.InputSpec((x.name for x in t.args), t.preds))
        return input_statement


def InputSpecParser():
    '''Syntax of the INPUT statement (supports comments starting with '%').'''
    with PyParsingDefaultWhitespaceChars(DEFAULT_WHITESPACE_CHARS):
        return ignore_comments(RawInputSpecParser())


def RawOutputSpecParser():
    '''Syntax of the OUTPUT statement (and nothing else).'''
    with PyParsingDefaultWhitespaceChars(DEFAULT_WHITESPACE_CHARS):
        OUTPUT = CaselessKeyword('OUTPUT').suppress()
        QUERY = CaselessKeyword('query').suppress()
        INDEX = CaselessKeyword('index').suppress()
        KEY = CaselessKeyword('key').suppress()
        CONTENT = CaselessKeyword('content').suppress()
        SET = CaselessKeyword('set').suppress()
        SEQUENCE = CaselessKeyword('sequence').suppress()
        DICTIONARY = CaselessKeyword('dictionary').suppress()
        NOT = CaselessKeyword('not').suppress()

        constant = integer | QuotedString('"', escChar='\\')
        constant.setParseAction(lambda t: o.Constant(t[0]))  # not strictly necessary to wrap this, but it simplifies working with the syntax tree

        asp_variable_name = Word(alphas_uppercase, alphanums + '_')
        asp_variable_anonymous = Keyword('_')
        asp_variable = asp_variable_anonymous | asp_variable_name
        asp_variable_expr = asp_variable_name.copy()
        #
        asp_variable_name.setParseAction(lambda t: asp.Variable(t[0]))
        asp_variable_anonymous.setParseAction(lambda t: asp.AnonymousVariable())
        asp_variable_expr.setParseAction(lambda t: o.Variable(t[0]))

        # TODO:
        # Instead of explicitly marking references with '&', we might just define a convention as follows:
        #   * Output names start with lowercase characters
        #   * ASP variables start with uppercase characters (as they do in actual ASP code)
        reference = amp + py_identifier
        reference.setParseAction(lambda t: o.Reference(t[0]))  # to distinguish from literal string values

        # Note: must be able to distinguish between unquoted and quoted constants
        asp_constant_symbol = Word(alphas_lowercase, alphanums + '_')
        asp_quoted_string = QuotedString('"', escChar='\\')
        asp_quoted_string.setParseAction(lambda t: asp.QuotedConstant(t[0]))
        term = (asp_constant_symbol | asp_quoted_string | asp_variable | positive_integer).setResultsName('terms', listAllMatches=True)
        terms = Optional(term + ZeroOrMore(comma + term))
        classical_atom = predicate_name('predicate') + Optional(lpar + terms + rpar)
        # Builtin atoms
        builtin_op_binary = (Literal('=') | '==' | '!=' | '<>' | '<' | '<=' | '>' | '>=' | '#succ').setResultsName('predicate')
        builtin_atom_binary = term + builtin_op_binary + term
        builtin_atom_binary_prefix = builtin_op_binary + lpar + term + comma + term + rpar
        builtin_atom = builtin_atom_binary | builtin_atom_binary_prefix
        #
        body_atom = classical_atom | builtin_atom
        pos_body_atom = body_atom.copy()
        neg_body_atom = NOT + body_atom
        pos_body_atom.setParseAction(lambda t: asp.Literal(t.predicate, tuple(t.terms), False))
        neg_body_atom.setParseAction(lambda t: asp.Literal(t.predicate, tuple(t.terms), True))
        body_literal = neg_body_atom | pos_body_atom
        #
        asp_query = Group(body_literal + ZeroOrMore(comma + body_literal))
        asp_query.setParseAction(lambda t: asp.Query(tuple(t[0])))

        expr = Forward()

        # TODO: Instead of semicolon, we could use (semicolon | FollowedBy(rbrace)) to make the last semicolon optional (but how would that work with asp_query...)
        query_clause = QUERY + colon + asp_query('query') + semicolon
        content_clause = CONTENT + colon + expr('content') + semicolon
        index_clause = INDEX + colon + asp_variable_expr('index') + semicolon
        key_clause = KEY + colon + expr('key') + semicolon
        #
        simple_set_spec = SET + lbrace + predicate_name('predicate') + slash + positive_integer('arity') + Optional(rightarrow + py_qualified_identifier('constructor')) + rbrace
        set_spec = SET + lbrace + (query_clause & content_clause) + rbrace
        # TODO: add clause like "at_missing_index: skip;", "at_missing_index: 0;", "at_missing_index: None;"
        sequence_spec = SEQUENCE + lbrace + (query_clause & content_clause & index_clause) + rbrace
        dictionary_spec = DICTIONARY + lbrace + (query_clause & content_clause & key_clause) + rbrace
        expr_collection = set_spec | simple_set_spec | sequence_spec | dictionary_spec
        #
        simple_set_spec.setParseAction(lambda t: o.ExprSimpleSet(t.predicate, t.arity, t.get('constructor')))
        set_spec.setParseAction(lambda t: o.ExprSet(t.query, t.content))
        sequence_spec.setParseAction(lambda t: o.ExprSequence(t.query, t.content, t.index))
        dictionary_spec.setParseAction(lambda t: o.ExprDictionary(t.query, t.content, t.key))

        expr_obj_args = Group(Optional(expr + ZeroOrMore(comma + expr) + Optional(comma)))
        expr_obj = Optional(py_qualified_identifier, default=None)('constructor') + lpar + expr_obj_args('args') + rpar
        #
        expr_obj.setParseAction(lambda t: o.ExprObject(t.constructor, t.args))

        # Note: "|" always takes the first match, that's why we have to parse variable names after obj (otherwise "variable name" might consume the identifier of expr_obj)
        expr << (constant | expr_collection | expr_obj | reference | asp_variable_expr)

        named_output_spec = py_identifier('name') + equals + expr('expr') + semicolon
        output_statement = OUTPUT + lbrace + ZeroOrMore(named_output_spec) + rbrace
        #
        named_output_spec.setParseAction(lambda t: (t.name, t.expr))
        output_statement.setParseAction(lambda t: o.OutputSpec(t))
        return output_statement


def OutputSpecParser():
    '''Syntax of the OUTPUT statement (supports comments starting with '%').'''
    with PyParsingDefaultWhitespaceChars(DEFAULT_WHITESPACE_CHARS):
        return ignore_comments(RawOutputSpecParser())


def RawSpecParser():
    '''Syntax of the whole I/O mapping specification: One INPUT statement and one OUTPUT statement in any order. This parser does not support comments.'''
    with PyParsingDefaultWhitespaceChars(DEFAULT_WHITESPACE_CHARS):
        i = RawInputSpecParser().setResultsName('input')
        o = RawOutputSpecParser().setResultsName('output')
        p = Optional(i) & Optional(o)
        # Note: t.get(n) returns None if n doesn't exist while t.n would return an empty string
        p.setParseAction(lambda t: (t.get('input'), t.get('output')))  # TODO
        return p


def SpecParser():
    '''Syntax of the whole I/O mapping specification: One INPUT statement and one OUTPUT statement in any order. This parser supports comments starting with '%'.'''
    with PyParsingDefaultWhitespaceChars(DEFAULT_WHITESPACE_CHARS):
        return ignore_comments(RawOutputSpecParser())


class EmbeddedSpecParser:
    '''Syntax of the whole I/O mapping specification, embedded in ASP comments starting with '%!'.'''
    # I tried doing this part with pyparsing too, so the whole parsing can be performed in a single pass without an intermediate string representation.
    # However, I was not able to make it work yet, so I am using a simple regex-based implementation at the moment.
    # Most likely problem with the pyparsing attempt: automatic handling of whitespace combined with LineStart() and LineEnd()
    # See also: http://pyparsing.wikispaces.com/share/view/18478063
    # The old attempt:
    #     p = SpecParser()
    #     asp_quoted_string = QuotedString('"', escChar='\\')
    #     asp_end = LineEnd() | '%!' | ('%' + ZeroOrMore(CharsNotIn('\n')) + LineEnd())       # '%!' must be before the regular comments since the '|' operator matches the first subexpression (MatchFirst)
    #     asp_line = LineStart() + ZeroOrMore(CharsNotIn('"%') | asp_quoted_string) + asp_end
    #     p.ignore(asp_line)
    #     return p
    #
    # This seems to work better, although neither LineStart() nor LineEnd() will match if the line starts with whitespace
    # (presumably because the parser by default skips as much whitespace as possible after parsing a token):
    #    asp_quoted_string = QuotedString('"', escChar='\\')
    #    asp_end = LineEnd().leaveWhitespace() | '%!' | ('%' + ZeroOrMore(CharsNotIn('\n')) + LineEnd().leaveWhitespace())       # '%!' must be before the regular comments since the '|' operator matches the first subexpression (MatchFirst)
    #    asp_line = (LineStart().leaveWhitespace() | LineEnd().leaveWhitespace()) + ZeroOrMore(CharsNotIn('"%\n') | asp_quoted_string) + asp_end
    #    p.ignore(asp_line)
    #
    # This seems to work quite well (implementation of comments inside %! part is still missing though, would be an ignore on SpecParser()):
    #    ParserElement.setDefaultWhitespaceChars(' \t\r')
    #    p = ZeroOrMore(Word(printables))
    #    # p.setWhitespaceChars(' \t\r')  # TODO: What to do with '\r'?
    #    linebreak = White('\n')
    #    asp_quoted_string = QuotedString('"', escChar='\\')
    #    asp_end = FollowedBy(linebreak) | '%!' | ('%' + ZeroOrMore(CharsNotIn('\n')) + FollowedBy(linebreak))       # '%!' must be before the regular comments since the '|' operator matches the first subexpression (MatchFirst)
    #    asp_line = (linebreak | StringStart()) + ZeroOrMore(CharsNotIn('"%\n') | asp_quoted_string) + asp_end
    #    p.ignore(asp_line)
    #    p.ignore(linebreak)

    spec_parser = RawSpecParser()

    # TODO:
    # A reasonable simplification might be to only allow %! comments for input specification at the start of a line,
    # i.e. only some whitespace may be before %! comments, and no ASP code.
    embedded_re = re.compile(r'''
        ^  # Start of each line (in MULTILINE mode)
        # The ASP part before comments
        (?:
            [^\n%"]  # anything except newlines, comment start, and quotes
            |
            # Quoted string: any char except newlines/backslash/quotes, or backslash escape sequences
            " (?: [^\n\\"] | \\. )* "
        )*
        %\!                     # Our specification language is embedded in special %! comments
        (?P<spec> [^\n%]* )     # The part we want to extract
        (?: %.* )?              # Comments in the specification language also start with % (like regular ASP comments)
        $  # end of each line (in MULTILINE mode)
    ''', re.MULTILINE | re.VERBOSE)

    @classmethod
    def extractFromString(cls, string):
        return '\n'.join(m.group('spec') for m in cls.embedded_re.finditer(string))

    def parseString(self, string, *, parseAll=True):
        return (_parse(type(self).spec_parser, type(self).extractFromString(string)),)


def AnswerSetParser():
    '''Parse the answer set from a single line of dlvhex2's output.'''
    with PyParsingDefaultWhitespaceChars(DEFAULT_WHITESPACE_CHARS):
        # As per the specification, we always return constants as strings. Conversion has to be performed explicitly with int(). See also `asp.RawAnswerSet` type.
        str_integer = Word(nums).setName('integer')
        quoted_string = QuotedString(quoteChar='"', escChar='\\')
        constant_symbol = Word(alphas_lowercase, alphanums + '_')
        arg = str_integer | quoted_string | constant_symbol
        fact = predicate_name('pred') + Group(Optional(lpar + arg + ZeroOrMore(comma + arg) + rpar))('args')
        answer_set = lbrace + Optional(fact + ZeroOrMore(comma + fact)) + rbrace  # + LineEnd()
        #
        fact.setParseAction(lambda t: (t.pred, tuple(t.args)))

        def collect_facts(t) -> asp.RawAnswerSet:
            d = {}  # type: MutableMapping[str, List[Tuple[str, ...]]]
            for (pred, args) in t:
                if pred not in d:
                    d[pred] = [args]
                    # Note:
                    # Technically we should use a set instead of a list here,
                    # but the ASP solver already performs the deduplication for us
                    # so there is no need to check for collisions again.
                    #
                    # Note:
                    # dlvhex2 differentiates between abc and "abc",
                    # but on the python side they are represented by the same string 'abc'.
                    # It is the responsibility of the ASP programmer to ensure quoted and
                    # non-quoted strings aren't mixed (this library only generates quoted
                    # strings during input mapping).
                    #
                    # What we could do:
                    # * Wrap unquoted constants in a class (so the "common" case of quoted constants is as before)
                    # * Or, probably better: Issue a warning if quoted and unquoted constants are mixed (maybe check this only in debug mode)
                else:
                    d[pred].append(args)
            return d  # type: ignore
        answer_set.setParseAction(collect_facts)
        return answer_set


def _parse(parser, string):
    try:
        result = parser.parseString(string, parseAll=True)
        return result[0]
    except ParseException:
        raise
    # except:
    #     pass  # TODO


class LazyInit:
    def __init__(self, constructor):
        self._lazy_constructor = constructor
        self._lazy_obj = None

    @property
    def lazy_obj(self):
        if self._lazy_obj is None:
            self._lazy_obj = self._lazy_constructor()
        return self._lazy_obj

    def __getattr__(self, name):
        return getattr(self.lazy_obj, name)


input_spec_parser = LazyInit(InputSpecParser)
output_spec_parser = LazyInit(OutputSpecParser)
spec_parser = LazyInit(SpecParser)
embedded_spec_parser = LazyInit(EmbeddedSpecParser)
answer_set_parser = LazyInit(AnswerSetParser)


def parse_input_spec(string):
    return _parse(input_spec_parser, string)


def parse_output_spec(string):
    return _parse(output_spec_parser, string)


def parse_spec(string):
    return _parse(spec_parser, string)


def parse_embedded_spec(string):
    return _parse(embedded_spec_parser, string)


def parse_answer_set(string: str) -> asp.RawAnswerSet:
    return _parse(answer_set_parser, string)



from abc import ABCMeta, abstractmethod
from contextlib import contextmanager
from itertools import chain
from typing import AbstractSet, Any, Iterable, Tuple, Optional, Union, Mapping, MutableMapping, Sequence  # noqa
from .errors import CircularReferenceError, DuplicateKeyError, InvalidIndicesError, RedefinedNameError, UndefinedNameError
from .registry import Registry
from . import asp
from . import parser

__all__ = [
    'OutputSpec'
]


class OutputResult:
    __object_is_being_mapped = object()  # sentinel value, used for cycle detection

    def __init__(self, toplevel: Mapping[str, 'Expr'], answer_set: asp.RawAnswerSet, registry: Registry) -> None:
        self.toplevel = toplevel
        self.answer_set = answer_set
        self.registry = registry
        # The objects created from the given answer set and registry
        self.objs = {}  # type: MutableMapping[str, Any]

    def get_object(self, name: str) -> Any:
        if name not in self.objs:
            if name not in self.toplevel:
                raise UndefinedNameError('No top-level name "{0}".'.format(name))
            self.objs[name] = OutputResult.__object_is_being_mapped
            self.objs[name] = self.toplevel[name].evaluate(self, LocalContext())
            # We don't need to keep the raw data around after everything has been mapped
            if len(self.toplevel) == len(self.objs):
                del self.answer_set
        if self.objs[name] is OutputResult.__object_is_being_mapped:
            raise CircularReferenceError('Circular reference detected while trying to resolve name "{0}".'.format(name))
        return self.objs[name]


class LocalContext:
    def __init__(self) -> None:
        # Current assignment of ASP variables
        self.va = {}  # type: MutableMapping[str, str]

    @contextmanager
    def assign_variables(self, names: Sequence[str], values: Sequence[str]):
        assert len(names) == len(values)
        for (name, value) in zip(names, values):
            assert name not in self.va
            self.va[name] = value
        yield
        for name in names:
            del self.va[name]


class Expr(metaclass=ABCMeta):
    @abstractmethod
    def evaluate(self, r: OutputResult, lc: LocalContext) -> Any:
        pass

    def variables(self) -> Iterable['Variable']:
        '''All the variable expressions (i.e., does not include those occurring only in queries) used in this expression and any subexpressions.'''
        return ()

    def check(self, toplevel_name: str, bound_variables: Tuple[str, ...]) -> None:
        pass

    def additional_rules(self) -> Iterable[asp.Rule]:
        return ()

    def captured_predicates(self) -> Iterable[str]:
        return ()


class Constant(Expr):
    def __init__(self, value: Union[int, str]) -> None:
        self.value = value

    def evaluate(self, r: OutputResult, lc: LocalContext) -> Union[int, str]:
        return self.value


class Reference(Expr):
    def __init__(self, name: str) -> None:
        assert len(name) > 0
        self.name = name

    def evaluate(self, r: OutputResult, lc: LocalContext) -> Any:
        return r.get_object(self.name)


class Variable(Expr):
    def __init__(self, name: str) -> None:
        assert len(name) > 0
        self.name = name

    def __str__(self):
        return self.name

    def __repr__(self):
        return 'Variable({0!r})'.format(self.name)

    def evaluate(self, r: OutputResult, lc: LocalContext) -> str:
        assert self.name in lc.va
        return str(lc.va[self.name])

    def variables(self) -> Iterable['Variable']:
        return [self]

    def check(self, toplevel_name: str, bound_variables: Tuple[str, ...]) -> None:
        if self.name not in bound_variables:
            raise UndefinedNameError('Variable {0!s} is not defined at point of use (at least once in definition of {1!s}).'.format(self.name, toplevel_name))


# TODO: We keep saying "constructor", but the implementation itself supports arbitrary functions, whether they construct a new object or not (but there are no guarantees w.r.t. side effects).
class ExprObject(Expr):
    def __init__(self, constructor_name: Optional[str], args: Iterable[Expr]) -> None:
        self.constructor_name = constructor_name  # If None, we will construct a tuple
        self.args = tuple(args)

    def evaluate(self, r: OutputResult, lc: LocalContext) -> Any:
        if self.constructor_name is None:
            def make_tuple(*args):
                # `tuple()` expects an Iterable, but the `constructor` is fed a number of separate arguments
                return r.registry.tuple_constructor(args)
            constructor = make_tuple
        else:
            constructor = r.registry.resolve(self.constructor_name)
        if constructor is None:
            raise NotImplementedError('constructor {0!r} not defined'.format(self.constructor_name))  # TODO
        mapped_args = (subexpr.evaluate(r, lc) for subexpr in self.args)
        return constructor(*mapped_args)

    def variables(self) -> Iterable['Variable']:
        # return chain(*(subexpr.variables() for subexpr in self.args))
        for subexpr in self.args:
            yield from subexpr.variables()

    def additional_rules(self) -> Iterable[asp.Rule]:
        for subexpr in self.args:
            yield from subexpr.additional_rules()
        # return chain(*(subexpr.additional_rules() for subexpr in self.args))

    def captured_predicates(self) -> Iterable[str]:
        for subexpr in self.args:
            yield from subexpr.captured_predicates()
        # return chain(*(subexpr.captured_predicates() for subexpr in self.args))

    def check(self, toplevel_name: str, bound_variables: Tuple[str, ...]) -> None:
        for subexpr in self.args:
            subexpr.check(toplevel_name, bound_variables)


def ExprSimpleSet(predicate_name: str, arity: int, constructor_name: Optional[str]) -> Expr:
    # Simple set semantics: just take the tuples as-is and put them in a set.
    varnames = ['X' + str(i) for i in range(arity)]
    literal = asp.Literal(predicate_name, tuple(asp.Variable(v) for v in varnames), False)
    query = asp.Query((literal,))
    if arity == 1 and constructor_name is None:
        # 1-tuples are unpacked automatically
        content = Variable(varnames[0])  # type: Expr
    else:
        content = ExprObject(constructor_name, tuple(Variable(v) for v in varnames))
    return ExprSet(query, content)


# TODO: Separate ABC ExprComposed for the subexpressions? Could then use that in ExprObject too
class ExprCollection(Expr):
    def __init__(self, query: asp.Query, subexpressions: Sequence[Expr]) -> None:
        self.query = query
        self.subexpressions = tuple(subexpressions)
        # # TODO: Note the precondition somewhere: no predicate starting with aspio__ may be used anywhere in the ASP program (or our additional rules will alter the program's meaning).
        # TODO: Special handling for cases where the original predicate can be used (e.g., only one literal and all variables captured)
        # TODO: Similar output expressions may generate duplicate rules... would be good if both expressions could use the same rule, without having all the data twice in the answer set
        self.output_predicate = 'aspio__' + str(id(self))  # unique for as long as this object is alive
        self.captured_variables = None  # type: Tuple[str, ...]

    def additional_rules(self) -> Iterable[asp.Rule]:
        rule = self.output_predicate + '(' + ','.join(self.captured_variables) + ') :- ' + str(self.query) + '.'
        yield rule
        for subexpr in self.subexpressions:
            yield from subexpr.additional_rules()
        # return chain([rule], *(expr.additional_rules() for expr in self.subexpressions))

    def captured_predicates(self) -> Iterable[str]:
        # return chain([self.output_predicate], *(expr.captured_predicates() for expr in self.subexpressions))
        yield self.output_predicate
        for subexpr in self.subexpressions:
            yield from subexpr.captured_predicates()

    def variables(self) -> Iterable[Variable]:
        for subexpr in self.subexpressions:
            yield from subexpr.variables()

    def check(self, toplevel_name: str, bound_variables: Tuple[str, ...]) -> None:
        # NOTE: We have to use the variable names (i.e., strings) here,
        #       because the query returns ASP variable objects, while the expressions return variable expression objects.
        #
        # All variables that appear in the query
        query_variables = set(str(v) for v in self.query.variables())
        # All variables that are fixed from the surrounding expression
        # (semantically equivalent: variables that are replaced by constants before evaluating the query for this expression)
        self.fixed_query_variables = tuple(query_variables.intersection(bound_variables))
        # All variables that are varying in the context of one result of this expression
        # (i.e., these variables vary and thus the content subexpression results in different contained objects),
        # and are also used in the construction of at least one subexpression
        used_varying_query_variables = query_variables.intersection(str(v) for v in self.variables())
        # We need to capture all fixed and (used) varying variables in the query, and ignore all others
        # IMPORTANT: The fixed variables must come first! (cf. get_captured_values)
        self.captured_variables = self.fixed_query_variables + tuple(used_varying_query_variables.difference(self.fixed_query_variables))
        assert len(set(self.captured_variables)) == len(self.captured_variables)
        self.used_varying_query_variables = self.captured_variables[len(self.fixed_query_variables):]
        assert set(self.used_varying_query_variables) == used_varying_query_variables

        for subexpr in self.subexpressions:
            subexpr.check(toplevel_name, bound_variables + tuple(query_variables))

    def get_captured_values(self, r: OutputResult, lc: LocalContext) -> Iterable[Tuple[str, ...]]:
        '''Return only those tuples of the `output_predicate` that assign the correct values for the fixed variables.'''
        for captured_values in r.answer_set.get(self.output_predicate, ()):
            # Only yield tuples that assign the correct values for the fixed variables
            for (i, v) in enumerate(self.fixed_query_variables):
                assert self.captured_variables[i] == v
                if lc.va[v] != captured_values[i]:
                    break
            else:
                # fixed_query_variables was exhausted without reaching `break`, which means all fixed variables have the correct value
                yield captured_values

    # @contextmanager
    # def assign_varying_variables(self, lc: LocalContext, values: Sequence[str]):
    #     assert len(self.captured_variables) == len(values)
    #     # Only assign the (used) varying values, the fixed variables are already assigned
    #     i = len(self.fixed_query_variables)
    #     with lc.assign_variables(self.captured_variables[i:], values[i:]):
    #         yield

    def eval_subexpression(self, expr: Expr, captured_values: Sequence[str], r: OutputResult, lc: LocalContext) -> Any:
        assert len(self.captured_variables) == len(captured_values)
        # Only assign the (used) varying values, the fixed variables are already assigned
        i = len(self.fixed_query_variables)
        with lc.assign_variables(self.captured_variables[i:], captured_values[i:]):
            return expr.evaluate(r, lc)


class ExprSet(ExprCollection):
    # TODO: Note the limitation somewhere (contents must be hashable)

    def __init__(self, query: asp.Query, content: Expr) -> None:
        super().__init__(query, [content])
        self.content = content

    def evaluate(self, r: OutputResult, lc: LocalContext) -> AbstractSet[Any]:
        make_set = r.registry.set_constructor  # type: ignore
        gen = (self.eval_subexpression(self.content, vs, r, lc) for vs in self.get_captured_values(r, lc))
        return make_set(gen)  # type: ignore


class ExprSequence(ExprCollection):
    def __init__(self, query: asp.Query, content: Expr, index: Variable) -> None:
        super().__init__(query, [content, index])
        self.content = content
        self.index = index

    def evaluate(self, r: OutputResult, lc: LocalContext) -> Sequence[Any]:
        def index_for(captured_values):
            str_value = captured_values[self.index_pos]
            try:
                return int(str_value)
            except ValueError as e:
                raise InvalidIndicesError('index variable is not an integer: {0!s}'.format(e))

        def content_for(captured_values):
            return self.eval_subexpression(self.content, captured_values, r, lc)

        # TODO: Options to determine how missing/duplicate indices should be handled
        # Currently: We require the indices to form a range of integers from 0 to n without any duplicates.
        all_captured_values = tuple(self.get_captured_values(r, lc))
        indices = sorted(index_for(captured_values) for captured_values in all_captured_values)
        if indices != list(range(len(indices))):
            raise InvalidIndicesError('not a valid range of indices')  # TODO: better message
        xs = sorted((index_for(captured_values), content_for(captured_values)) for captured_values in all_captured_values)
        make_sequence = r.registry.sequence_constructor  # type: ignore
        return make_sequence(x[1] for x in xs)  # type: ignore

    def check(self, toplevel_name: str, bound_variables: Tuple[str, ...]) -> None:
        super().check(toplevel_name, bound_variables)
        self.index_pos = self.captured_variables.index(self.index.name)


class ExprDictionary(ExprCollection):
    # TODO: Note the limitation somewhere (keys must be hashable)

    def __init__(self, query: asp.Query, content: Expr, key: Expr) -> None:
        super().__init__(query, [content, key])
        self.content = content
        self.key = key

    def evaluate(self, r: OutputResult, lc: LocalContext) -> Mapping[Any, Any]:
        d = {}  # type: MutableMapping[Any, Any]
        for captured_values in self.get_captured_values(r, lc):
            k = self.eval_subexpression(self.key, captured_values, r, lc)
            if k not in d:
                d[k] = self.eval_subexpression(self.content, captured_values, r, lc)
            else:
                raise DuplicateKeyError('Duplicate key: {0}'.format(repr(k)))
        make_dictionary = r.registry.dictionary_constructor  # type: ignore
        return make_dictionary(d)  # type: ignore


class OutputSpec:
    def __init__(self, named_exprs: Iterable[Tuple[str, Expr]]) -> None:
        exprs = {}  # type: MutableMapping[str, Expr]
        for (name, expr) in named_exprs:
            if name not in exprs:
                exprs[name] = expr
            else:
                raise RedefinedNameError('Duplicate top-level name: {0}'.format(name))
        # Note: easier with dict(named_exprs), check len(exprs) == len(named_exprs); but: error message is not as meaningful!
        self.exprs = exprs  # type: Mapping[str, Expr]
        # TODO: Check for cycles in references (currently we do that while mapping, but for consistency it would be nice to have it checked at time of construction -- it is some additional work though, while we get the result 'for free' during mapping)
        for (name, expr) in self.exprs.items():
            expr.check(toplevel_name=name, bound_variables=())  # , bound_references=self.exprs.keys())

    @staticmethod
    def empty() -> 'OutputSpec':
        return OutputSpec(())

    @staticmethod
    def parse(string: str) -> 'OutputSpec':
        return parser.parse_output_spec(string)

    def prepare_mapping(self, answer_set: asp.RawAnswerSet, registry: Registry) -> OutputResult:
        return OutputResult(self.exprs, answer_set, registry)

    def additional_rules(self) -> Iterable[asp.Rule]:
        for expr in self.exprs.values():
            yield from expr.additional_rules()

    def captured_predicates(self) -> Iterable[str]:
        # create a set to remove duplicates
        return set(chain(*(expr.captured_predicates() for expr in self.exprs.values())))



import io
import signal
import subprocess  # type: ignore
import weakref
from itertools import chain
from typing import Callable, IO, Iterable, Iterator, Optional
from ..helper.typing import ClosableIterable
from ..errors import SolverError, SolverSubprocessError
from ..helper import FilesystemIPC, StreamCaptureThread, TemporaryFile, TemporaryNamedPipe
from ..parser import parse_answer_set, ParseException
from .abc import Solver, SolverOptions
from .. import asp


class Dlvhex2Solver(Solver):
    '''Interface to the dlvhex2 solver.'''

    # TODO: Check what encoding dlvhex2 expects (on stdin and for input files) -- this is not supposed to be an option, but the encoding that dlvhex2 uses to read from stdin (and input files)
    default_encoding = 'UTF-8'

    def __init__(self, *, executable: str = None) -> None:
        '''Initialize a dlvhex2 solver instance.

        @param executable The path to the dlvhex2 executable. If not specified, looks for "dlvhex2" in the current $PATH.
        '''
        self.executable = executable if executable is not None else 'dlvhex2'
        self.encoding = type(self).default_encoding

    def __copy__(self) -> 'Solver':
        other = Dlvhex2Solver()
        other.executable = self.executable
        other.encoding = self.encoding
        return other

    def run(self, *,
            write_input: Callable[[IO[str]], None],
            capture_predicates: Iterable[str],
            file_args: Iterable[str],
            options: Optional[SolverOptions]) -> ClosableIterable[asp.RawAnswerSet]:
        '''Run the dlvhex solver on the given program.'''
        # Prefer named pipes, but fall back to a file if pipes are not implemented for the current platform
        try:
            tmp_input = TemporaryNamedPipe()  # type: FilesystemIPC
        except NotImplementedError:
            tmp_input = TemporaryFile()

        if options is not None and options.capture is not None:
            capture_predicates = chain(capture_predicates, options.capture)

        try:
            args = [
                self.executable,
                # only print the answer sets themselves
                '--silent',
                # only capture relevant predicates
                '--filter=' + ','.join(capture_predicates),
                # wait for a newline on stdin between answer sets
                '--waitonmodel',
            ]
            # options passed in by caller
            if options is not None:
                if options.max_answer_sets is not None:
                    args.append('--number={0!s}'.format(options.max_answer_sets))
                if options.max_int is not None:
                    args.append('--maxint={0!s}'.format(options.max_int))
                if options.custom is not None:
                    args.extend(options.custom)
            # tell dlvhex2 to read our input from the named pipe
            args.append(tmp_input.name)
            args.extend(file_args)

            # If we have a temporary file, we must pass the data before starting the subprocess
            if isinstance(tmp_input, TemporaryFile):
                with open(tmp_input.name, 'wt', encoding=self.encoding) as stream:
                    write_input(stream)

            # Start dlvhex2 subprocess
            process = subprocess.Popen(
                args,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

            try:
                # If we have a named pipe, we must pass the data after starting the subprocess,
                # or we risk a deadlock by filling the pipe's buffer
                if isinstance(tmp_input, TemporaryNamedPipe):
                    with open(tmp_input.name, 'wt', encoding=self.encoding) as stream:
                        write_input(stream)
                    # At this point the input pipe is flushed and closed, and dlvhex2 starts processing

                lines = DlvhexLineReader(process=process, encoding=self.encoding, tmp_input=tmp_input)
                return AnswerSetParserIterable(lines)
            except:
                process.kill()
                # Close streams to prevent ResourceWarnings
                process.stdin.close()
                process.stdout.close()
                process.stderr.close()
                raise
        except:
            tmp_input.cleanup()
            raise


class DlvhexLineReader(ClosableIterable[str]):
    '''Wraps a process and provides its standard output for line-based iteration.

    It is only possible to iterate *once* over a DlvhexLineReader instance.

    If the process exits with a return code other than 0,
    a SolverSubprocessError will be thrown during iteration, containing the return code and stderr output of the process.
    '''

    def __init__(self, *, process: subprocess.Popen, encoding: str, tmp_input: FilesystemIPC) -> None:
        self.process = process
        self.stdout_encoding = encoding
        self.iterating = False
        #
        # We need to capture stderr in a background thread to avoid deadlocks.
        # (The problem would occur when dlvhex2 is blocked because the OS buffers on the stderr pipe are full... so we have to constantly read from *both* stdout and stderr)
        self.stderr_capture_thread = StreamCaptureThread(self.process.stderr)
        self.stderr_capture_thread.start()
        #
        # Set up finalization. Using weakref.finalize seems to work more robustly than using __del__.
        # (One problem that occurred with __del__: It seemed like python was calling __del__ for self.process and its IO streams first,
        # which resulted in ResourceWarnings even though we were closing the streams properly in our __del__ function.)
        self._finalize = weakref.finalize(self, DlvhexLineReader.__close, process, self.stderr_capture_thread, encoding, tmp_input)  # type: ignore
        # Make sure the subprocess will be terminated if it's still running when the python process exits
        self._finalize.atexit = True

    def __iter__(self) -> Iterator[str]:
        '''Return an iterator over the lines written to stdout. May only be called once! Might raise a SolverSubprocessError.'''
        assert not self.iterating, 'You may only iterate once over a single DlvhexLineReader instance.'
        self.iterating = True
        # Requirement: dlvhex2 needs to flush stdout after every line
        with io.TextIOWrapper(self.process.stdout, encoding=self.stdout_encoding) as stdout_lines:
            for line in stdout_lines:
                yield line
                # Tell dlvhex2 to prepare the next answer set
                if not self.process.stdin.closed:
                    self.process.stdin.write(b'\n')
                    self.process.stdin.flush()
                else:
                    break
        # We've exhausted stdout, so either:
        #   1. we got all answer sets, or
        #   2. an error occurred,
        # and dlvhex closed stdout (and probably terminated).
        # Give it a chance to terminate gracefully.
        try:
            self.process.wait(timeout=0.005)  # type: ignore (mypy does not know about `timeout`)
        except subprocess.TimeoutExpired:  # type: ignore (mypy does not know about `TimeoutExpired`)
            pass
        self.close()

    def close(self) -> None:
        self._finalize()

    # We cannot have a reference to `self` because we must avoid reference cycles here (see weakref.finalize documentation).
    @staticmethod
    def __close(process: subprocess.Popen, stderr_capture_thread: StreamCaptureThread[bytes], stderr_encoding: str, tmp_input: FilesystemIPC) -> None:
        '''Shut down the process if it is still running. Raise a SolverSubprocessError if the process exited with an error.'''
        if process.poll() is None:
            # Still running? Kill the subprocess
            process.terminate()
            try:
                process.wait(timeout=0.001)  # type: ignore (mypy does not know about `timeout`)
            except subprocess.TimeoutExpired:  # type: ignore (mypy does not know about `TimeoutExpired`)
                # Kill unconditionally after a short timeout.
                # A potential problem with SIGKILL: we might not get all the error messages on stderr (if the child process is killed before it has a chance to write an error message)
                process.kill()
                process.wait()
            assert process.poll() is not None  # subprocess has stopped
            # Various outcomes were observed that do not signify a real error condition as far as py-dlvhex is concerned:
            # * dlvhex2 simply terminates by itself between the is_running() check and the terminate() call
            # * dlvhex2 executes its SIGTERM handler and exits with code 2, and a message on stderr ("dlvhex2 [...] got termination signal 15")
            # * dlvhex2 doesn't have its SIGTERM handler active and the default handler exits with code -15
            # * dlvhex2 hangs after receiving SIGTERM (maybe when it's blocking on stdout? we're not reading from it anymore at this point), so we send SIGKILL after the timeout and it exits with -9
            # * dlvhex2 receives SIGTERM and crashes with "Assertion failed: (!res), function ~mutex, file /usr/local/include/boost/thread/pthread/mutex.hpp, line 111", and exit code -6 (SIGABRT)
            if process.returncode in (2, -signal.SIGTERM, -signal.SIGKILL, -signal.SIGABRT):
                # In the cases mentioned above (and only if we are responsible for termination, i.e. after calling terminate() from this function), don't throw an exception
                process.returncode = 0
        # Note: only close stdin after the process has been terminated, otherwise dlvhex will start outputting everything at once
        process.stdin.close()
        process.stdout.close()
        stderr_capture_thread.join()  # ensure cleanup of stderr
        # Remove the temporary input pipe/file
        # Note: To clean up tmp_input in a reliable way, we must be sure dlvhex2 has already read all the data (or, has at the very least opened the file).
        #       Because of this, the earliest point where we are able to clean up tmp_input would be when dlvhex2 starts outputting the first answer set.
        tmp_input.cleanup()
        if process.returncode != 0:
            err = SolverSubprocessError(process.returncode, str(stderr_capture_thread.data, encoding=stderr_encoding))
            process.returncode = 0  # make sure we only raise an error once
            raise err


class AnswerSetParserIterable(ClosableIterable[asp.RawAnswerSet]):
    def __init__(self, lines: ClosableIterable[str]) -> None:
        self.lines = lines

    # @staticmethod
    # def _parse(line):
    #     try:
    #         return parse_answer_set(line)
    #     except ParseException:
    #         e = SolverError('Unable to parse answer set received from solver')
    #         e.line = line  # type: ignore
    #         raise e

    def __iter__(self) -> Iterator[asp.RawAnswerSet]:
        # return iter(map(type(self)._parse, self.lines))
        for line in self.lines:
            try:
                yield parse_answer_set(line)
            except ParseException:
                e = SolverError('Unable to parse answer set received from solver')
                e.line = line  # type: ignore
                raise e

    def close(self) -> None:
        self.lines.close()



from threading import Thread
from typing import Generic, IO, TypeVar

__all__ = ['StreamCaptureThread']


S = TypeVar('S', str, bytes)


class StreamCaptureThread(Generic[S], Thread):
    '''A thread that reads all contents of the given stream.

    The stream's contents can be accessed through the data attribute after the thread has finished.
    '''

    def __init__(self, stream: IO[S]) -> None:
        super().__init__(daemon=True)
        self.stream = stream
        self.data = None  # type: S

    def run(self):
        '''Reads the stream's contents and closes the stream.'''
        with self.stream as s:
            self.data = s.read()



from abc import ABC, abstractmethod
from copy import copy
from typing import Callable, IO, Iterable, Optional, Sequence  # noqa
from ..helper.typing import ClosableIterable
from .. import asp

__all__ = [
    'Solver',
    'SolverOptions',
]


class SolverOptions:
    def __init__(self, *,
                 max_answer_sets: Optional[int] = None,
                 max_int: Optional[int] = None,
                 capture: Optional[Iterable[str]] = None,
                 custom: Optional[Sequence[str]] = None) -> None:
        self.max_answer_sets = max_answer_sets
        '''Instruct the solver to compute at most `max_answer_sets` answer sets. Compute all answer sets if `None`.'''
        self.max_int = max_int
        '''Set maximum integer value.'''
        self.capture = capture
        '''Capture additional predicates if they are to be examined manually.'''
        # TODO: Provide some "sentinel" value for the capture options that just means "capture everything"
        self.custom = custom
        '''Custom solver options, passed to the solver as-is'''
        # TODO: Add a "timeout" option? => creates a watchdog thread that just kills the solver after the time elapses (prompting a SolverTimeoutExpired exception or something like that.)

    def __copy__(self) -> 'SolverOptions':
        return SolverOptions(
            max_answer_sets=self.max_answer_sets,
            max_int=self.max_int,
            capture=copy(self.capture),
            custom=copy(self.custom))


class Solver(ABC):
    '''Abstract solver interface.'''

    @abstractmethod
    def run(self, *,
            write_input: Callable[[IO[str]], None],
            capture_predicates: Iterable[str],
            file_args: Iterable[str],
            options: SolverOptions = None) -> ClosableIterable[asp.RawAnswerSet]:
        pass

    @abstractmethod
    def __copy__(self) -> 'Solver':
        pass



import importlib
import logging
from copy import copy
from types import ModuleType
from typing import Any, AbstractSet, Callable, Iterable, Mapping, MutableMapping, Optional, Sequence, Union  # noqa

__all__ = [
    'Constructor',
    'global_registry',
    'import_from_module',
    'register',
    'register_dict',
    'Registry',
]

log = logging.getLogger(__name__)

Constructor = Callable[..., object]


class Registry:
    # Default constructors for collections (global setting)
    tuple_constructor = tuple  # type: Callable[[Iterable[Any]], object]
    set_constructor = frozenset  # type: Callable[[Iterable[Any]], AbstractSet[Any]]
    sequence_constructor = list  # type: Callable[[Iterable[Any]], Sequence[Any]]
    dictionary_constructor = dict  # type: Callable[[Mapping[Any, Any]], Mapping[Any, Any]]

    def __init__(self) -> None:
        self._registered_names = {
            'int': int  # 'int' constructor must be available as per the language specification
        }  # type: MutableMapping[str, Constructor]
        # Default constructors for collections (local setting)
        self.tuple_constructor = type(self).tuple_constructor  # type: ignore  # bug in mypy: https://github.com/python/mypy/issues/708
        self.set_constructor = type(self).set_constructor  # type: ignore
        self.sequence_constructor = type(self).sequence_constructor  # type: ignore
        self.dictionary_constructor = type(self).dictionary_constructor  # type: ignore

    def __copy__(self) -> 'Registry':
        other = Registry()
        other._registered_names = copy(self._registered_names)
        return other

    def register(self, constructor: Constructor, name: str = None, *, replace: bool = False) -> None:
        '''Register the given constructor with the given name.

        If `name` is not given, it defaults to `constructor.__name__` (raising a `ValueError` if this attribute does not exist).

        Raises a `ValueError` when trying to re-register a name with a different constructor (unless `replace` is `True`).
        Raises a `ValueError` when the constructor argument is not callable.
        '''
        if name is None:
            try:
                name = constructor.__name__
            except AttributeError:
                raise ValueError("Constructor {0!r} has no attribute '__name__'. Please provide the name argument manually.".format(constructor))
        if not replace and name in self._registered_names:
            if self._registered_names[name] is constructor:
                # If we try to register the same object again, there is no problem.
                # This might happen in practice when calling `register_dict` with `globals()`, since all the imported names are registered too.
                return
            raise ValueError('Name {0!r} is already registered. Pass replace=True to re-register.'.format(name))
        if not callable(constructor):
            raise ValueError('The constructor argument must be callable.')
        log.debug('Registry: registering name %r with constructor %r', name, constructor)
        self._registered_names[name] = constructor

    def register_dict(self, name_dict: Mapping[str, Any]) -> None:
        '''Import names from the given dict.

        Skips entries where the associated constructor is not callable.
        Skips entries with a 'special attribute'-like name (i.e., starting and ending with double underscores).
        Can be used to easily add all definitions of the current module when called as `register_dict(globals())`.
        '''
        for name, obj in name_dict.items():
            # Note: In practice, objects bound to these "special" names aren't callable anyways...
            # But the names are the same in every module, so we want to prevent conflicts in case one of them *does* become callable in a future version.
            if name.startswith('__') and name.endswith('__'):
                continue
            if callable(obj):
                self.register(obj, name)

    def import_from_module(self, names: Iterable[str], module_or_module_name: Union[ModuleType, str], package: Optional[str] = None) -> None:
        '''Import names from the given module.

        All objects bound to the given names in the module are registered.
        The state at the time of import is captured,
        i.e. no assignments to the given names after the import_from_module call will be noticed.
        '''
        if isinstance(module_or_module_name, ModuleType):
            module = module_or_module_name
        else:
            module = importlib.import_module(module_or_module_name, package=package)
        for name in names:
            self.register(getattr(module, name), name)

    def get(self, name: str) -> Constructor:
        '''Return the constructor registered to `name`, or `None` if the name is not registered.'''
        return self._registered_names.get(name)

    def resolve(self, name: str) -> Constructor:
        '''Resolve the given (possibly qualified) name.

        The name is split into parts separated by dots.
        The leftmost (top level) name is looked up in this registry or, if it is not registered, imported as module.
        '''
        toplevel, *parts = name.split('.')
        obj = self.get(toplevel)  # type: Any
        if obj is None:
            try:
                obj = importlib.import_module(toplevel)
            except ImportError:
                pass
        if obj is None:
            return None
        for part in parts:
            try:
                obj = getattr(obj, part)
            except AttributeError:
                return None
        if callable(obj):
            return obj
        else:
            return None


global_registry = Registry()

register = global_registry.register
register_dict = global_registry.register_dict
import_from_module = global_registry.import_from_module



import os
import tempfile
import warnings
import weakref
from abc import ABC, abstractmethod

__all__ = [
    'FilesystemIPC',
    'TemporaryFile',
    'TemporaryNamedPipe',
]


class FilesystemIPC(ABC):
    '''An IPC mechanism that is accessible via the filesystem.'''

    def __init__(self) -> None:
        self.name = None  # type: str

    @abstractmethod
    def cleanup(self) -> None:
        pass

    def __enter__(self) -> str:
        return self.name

    def __exit__(self, exc_type, exc_value, traceback):
        self.cleanup()
        return False


def _warn_cleanup(typename, filename):
    warnings.warn('Did not call cleanup() on {0} with name {1}'.format(typename, filename), ResourceWarning)


class TemporaryNamedPipe(FilesystemIPC):
    def __init__(self) -> None:
        '''Create a temporary named pipe.

        The path to the named pipe can be retrieved from the `name` attribute on the returned object.

        To ensure proper removal of the pipe after use,
        make sure to call `cleanup()` on the returned object, or use it as a context manager.

        Raises `OSError` if the named pipe could not be created.
        Raises `NotImplementedError` if `mkfifo` from the builtin `os` module is not available.
        '''
        if hasattr(os, 'mkfifo'):
            self.tmpdir = tempfile.TemporaryDirectory(prefix='pyaspio_')  # creates a temporary directory, avoiding any race conditions
            self.name = os.path.join(self.tmpdir.name, 'pipe')
            os.mkfifo(self.name)  # may raise OSError

            self._cleanup_warning = weakref.finalize(self, _warn_cleanup, type(self).__name__, self.name)  # type: ignore
            self._cleanup_warning.atexit = True
        else:
            # Notes about a possible Windows implementation:
            #   Windows provides named pipes, see the CreateNamedPipe function in the Windows API: https://msdn.microsoft.com/en-us/library/aa365150(v=vs.85).aspx
            #   We can use the `ctypes` module to call this function, e.g.:
            #
            #       import ctypes
            #       kernel32 = ctypes.windll.kernel32
            #       pipe_handle = kernel32.CreateNamedPipeW(name, ...)
            #
            #   Related functions: ConnectNamedPipe, WriteFile, DisconnectNamedPipe, CloseHandle.
            #
            #   Overall the Windows pipe system seems more complicated,
            #   and I was not yet able to make it work without also changing the client code (which expects to read from a file).
            #
            #   For now, fall back to a temporary file on Windows.
            raise NotImplementedError()

    def cleanup(self) -> None:
        if self._cleanup_warning.peek() is not None:
            self.tmpdir.cleanup()
            self._cleanup_warning.detach()


class TemporaryFile(FilesystemIPC):
    def __init__(self) -> None:
        '''Create a temporary file.

        The path to the file can be retrieved from the `name` attribute on the returned object.

        To ensure proper removal of the file after use,
        make sure to call `cleanup()` on the returned object, or use it as a context manager.
        '''
        fd, self.name = tempfile.mkstemp(prefix='pyaspio_')
        # Ideally, we would use this file descriptor instead of reopening the file from the path later,
        # but then we do not know whether the file descriptor has already been closed.
        # (possible problems: we might leak a file descriptor, or close it twice (thus risking to close another unrelated file))
        os.close(fd)

        self._cleanup_warning = weakref.finalize(self, _warn_cleanup, type(self).__name__, self.name)  # type: ignore
        self._cleanup_warning.atexit = True

    def cleanup(self) -> None:
        if self._cleanup_warning.peek() is not None:
            os.remove(self.name)
            self._cleanup_warning.detach()



__all__ = [
    'CircularReferenceError',
    'DuplicateKeyError',
    'InvalidIndicesError',
    'RedefinedNameError',
    'UndefinedNameError',
    'SolverError',
    'SolverSubprocessError',
]


class CircularReferenceError(ValueError):
    '''Raised when a circular reference is detected in an output specification.'''


class DuplicateKeyError(ValueError):
    '''Raised when duplicate keys appear while mapping a predicate to a dictionary.'''


class InvalidIndicesError(ValueError):
    '''Raised when the indices do not form a valid range of integers (i.e., there a duplicates or missing indices) while mapping a predicate to a list.'''


class RedefinedNameError(ValueError):
    '''Raised when a name is declared when it has already been declared in the surrounding scope.'''


class UndefinedNameError(ValueError):
    '''Raised when a name is referenced that is not bound at that location.'''


class SolverError(Exception):
    '''Raised when an error with the solver occurs.'''


class SolverSubprocessError(SolverError):
    '''Raised when the solver subprocess exits abnormally.'''
    def __init__(self, returncode, stderr):
        message = 'The ASP solver terminated with return code {0}.\nOutput on stderr:\n{1}'.format(returncode, stderr)
        super().__init__(message)
        self.returncode = returncode
        self.stderr = stderr



from .caching_iterable import CachingIterable
from .filesystem_ipc import FilesystemIPC, TemporaryNamedPipe, TemporaryFile
from .stream_capture_thread import StreamCaptureThread

__all__ = [
    'CachingIterable',
    #
    'FilesystemIPC',
    'TemporaryFile',
    'TemporaryNamedPipe',
    #
    'StreamCaptureThread',
]



import collections
import collections.abc  # type: ignore (mypy does not know about collections.abc)
from abc import ABCMeta, abstractmethod
from typing import Iterable, Any, Union, Dict, Iterator, MutableSet, Sequence, AbstractSet
from . import parser
from .errors import RedefinedNameError, UndefinedNameError

Context = Dict['Variable', Any]


# Maybe have perform_mapping return a Generator (with "yield Fact(pred, args)" or some such) instead of using FactAccumulator instances.
class FactAccumulator(metaclass=ABCMeta):
    @abstractmethod
    def add_fact(self, predicate: str, args: Sequence[Any]) -> None:
        pass


class AssignmentTarget(metaclass=ABCMeta):
    @abstractmethod
    def check_and_update_variable_bindings(self, bound_variables: MutableSet['Variable']) -> None:
        pass

    @abstractmethod
    def assign(self, value: Any, context: Context) -> None:
        pass


class AnonymousVariable(AssignmentTarget):
    def check_and_update_variable_bindings(self, bound_variables: MutableSet['Variable']) -> None:
        pass

    def assign(self, value: Any, context: Context) -> None:
        pass

    def __repr__(self):
        return 'AnonymousVariable()'

    def __str__(self):
        return '_'


class Variable(AssignmentTarget):
    def __init__(self, name: str) -> None:
        assert len(name) > 0
        self._name = name

    def assign(self, value: Any, context: Context) -> None:
        context[self] = value

    def check_and_update_variable_bindings(self, bound_variables: MutableSet['Variable']) -> None:
        if self in bound_variables:
            raise RedefinedNameError('Variable {0!s} is defined twice'.format(self))
        bound_variables.add(self)

    def __repr__(self):
        return 'Variable({0!r})'.format(self._name)

    def __str__(self):
        return self._name

    def __eq__(self, other):
        if isinstance(other, Variable):
            return self._name == other._name
        else:
            return NotImplemented

    def __hash__(self):
        return hash(self._name)


class TupleMatch(AssignmentTarget):
    def __init__(self, targets: Sequence[AssignmentTarget]) -> None:
        self._targets = tuple(targets)

    def assign(self, value: Any, context: Context) -> None:
        if len(value) != len(self._targets):
            raise ValueError('length mismatch')  # TODO better message
        for t, v in zip(self._targets, value):
            t.assign(v, context)

    def check_and_update_variable_bindings(self, bound_variables: MutableSet[Variable]) -> None:
        for t in self._targets:
            t.check_and_update_variable_bindings(bound_variables)

    def __repr__(self):
        return 'TupleMatch([{0}])'.format(','.join(repr(t) for t in self._targets))

    def __str__(self):
        return '({0})'.format(', '.join(str(t) for t in self._targets))


class Attribute:
    def __init__(self, name: str) -> None:
        self._name = name

    def access(self, obj: Any) -> Any:
        try:
            return getattr(obj, self._name)
        except AttributeError:
            # Raise a ValueError, since this situation occurs when the user passes wrong input arguments to the program
            raise ValueError('Unable to access attribute {0!r} on object {1!r} during INPUT mapping'.format(self._name, obj))

    def __str__(self) -> str:
        return '.' + str(self._name)


class Subscript:
    def __init__(self, key: object) -> None:
        self._key = key

    def access(self, obj: Any) -> Any:
        try:
            return obj[self._key]
        except (KeyError, IndexError, TypeError):
            # Raise a ValueError, since this situation occurs when the user passes wrong input arguments to the program
            raise ValueError('Unable to access subscript [{0!r}] on object {1!r} during INPUT mapping'.format(self._key, obj))

    def __str__(self) -> str:
        return '[{0!r}]'.format(self._key)


class Accessor:
    def __init__(self, variable: Variable, path: Sequence[Union[Attribute, Subscript]]) -> None:
        self._variable = variable
        self._path = tuple(path)

    def __str__(self) -> str:
        return str(self._variable) + ''.join(map(str, self._path))

    def check_variable_bindings(self, bound_variables: AbstractSet[Variable]) -> None:
        if self._variable not in bound_variables:
            raise UndefinedNameError('Undefined variable {0!s} is being accessed'.format(self._variable))

    def perform_access(self, context: Context) -> Any:
        '''Performs the represented object access relative to the given context.'''
        result = context[self._variable]
        for attr in self._path:
            result = attr.access(result)
        return result


class Iteration:
    def __init__(self, target: AssignmentTarget, accessor: Accessor) -> None:
        self._target = target
        self._accessor = accessor

    def check_and_update_variable_bindings(self, bound_variables: MutableSet[Variable]) -> None:
        self._accessor.check_variable_bindings(bound_variables)
        self._target.check_and_update_variable_bindings(bound_variables)

    def get_collection_iterator(self, context: Context) -> Iterator[Any]:
        collection = self._accessor.perform_access(context)
        if isinstance(collection, collections.abc.Set):  # type: ignore (mypy does not know about collections.abc)
            return iter(collection)
        elif isinstance(collection, collections.abc.Sequence):  # type: ignore (mypy does not know about collections.abc)
            return enumerate(collection)  # yields (index, element) tuples
        elif isinstance(collection, collections.abc.Mapping):  # type: ignore (mypy does not know about collections.abc)
            return iter(collection.items())  # yields (key, element) tuples
        else:
            raise ValueError(
                'During iteration {0!r}: '
                'collection object of type {1!r} could not be identified as Set, Sequence or Dictionary. '
                'It should inherit from collections.abc.Set, collections.abc.Sequence, or collections.abc.Mapping.'
                .format(self, type(collection)))

    def assign_to_target(self, value: Any, context: Context) -> None:
        self._target.assign(value, context)

    def __str__(self) -> str:
        return 'FOR {0!s} IN {1!s}'.format(self._target, self._accessor)


class Predicate:
    def __init__(self, predicate: str, arguments: Sequence[Accessor], iterations: Sequence[Iteration]) -> None:
        '''Represents a single input mapping that produces facts of a single predicate.

        @param predicate: The name of the predicate to generate. Must be a non-empty string.
        @param arguments: A list of Accessor instances that describe the arguments of the generated facts.
        @param iterations: A list of Iteration instances that describe how to generate the set of facts from the input arguments.
        '''
        self._predicate = predicate
        self._arguments = tuple(arguments)
        self._iterations = tuple(iterations)

    def check_variable_bindings(self, input_variables: Iterable[Variable]) -> None:
        bound_variables = set(input_variables)
        for it in self._iterations:
            it.check_and_update_variable_bindings(bound_variables)  # adds newly bound variables to the set
        for arg in self._arguments:
            arg.check_variable_bindings(bound_variables)

    def perform_mapping(self, initial_context: Context, accumulator: FactAccumulator) -> None:
        # TODO: Add some explanation and use better variable names (it/iter??? maybe rename InputIteration to InputLoop so we don't confuse the python iterator concept with our iteration concept)
        context = initial_context.copy()
        iter_stack = []  # type: List[Iterator[Any]]
        while True:
            # Advance innermost iteration
            if len(iter_stack) > 0:
                col_iter = iter_stack[-1]
                it = self._iterations[len(iter_stack) - 1]
                try:
                    it.assign_to_target(next(col_iter), context)
                except StopIteration:
                    iter_stack.pop()
                    if len(iter_stack) == 0:
                        # Last iteration stopped? We're done
                        break
                    else:
                        continue

            if len(iter_stack) == len(self._iterations):
                # All iterations have been performed
                # => Generate a fact
                accumulator.add_fact(self._predicate, tuple(arg.perform_access(context) for arg in self._arguments))
                if len(self._iterations) == 0:
                    break
            else:
                # Haven't yet performed all iterations, move on to inner iteration
                assert(len(iter_stack) < len(self._iterations))
                it = self._iterations[len(iter_stack)]
                iter_stack.append(it.get_collection_iterator(context))


class InputSpec:
    def __init__(self, parameters: Sequence[Variable], predicates: Iterable[Predicate]) -> None:
        '''Represents an INPUT statement, i.e. the complete input mapping description for an ASP program.

        @param parameters: The input parameters that need to be given when solving the program.
        @param predicates: A list of Predicate instances describing how to generate facts from the input arguments.
        '''
        self._parameters = tuple(parameters)
        self._predicates = tuple(predicates)
        # Check for name errors in input arguments (i.e. there must not be duplicate names)
        if len(self._parameters) != len(set(self._parameters)):
            raise RedefinedNameError('Input parameters must have unique names')
        # Check for name errors in accessor and iteration definitions (two kinds of errors: either using an undefined variable name, or redefining a variable name)
        for pred in self._predicates:
            pred.check_variable_bindings(self._parameters)

    @staticmethod
    def empty() -> 'InputSpec':
        return InputSpec((), ())

    @staticmethod
    def parse(string: str) -> 'InputSpec':
        return parser.parse_input_spec(string)

    def perform_mapping(self, arguments: Sequence[Any], accumulator: FactAccumulator) -> None:
        '''Perform the input mapping.

        Transforms the arguments to an ASP representation according to the InputSpec,
        and passes the results to the given accumulator (see FactAccumulator class).
        '''
        if len(arguments) != len(self._parameters):
            raise ValueError('Wrong number of arguments: expecting {0}, got {1}'.format(len(self._parameters), len(arguments)))
        for pred in self._predicates:
            context = dict(zip(self._parameters, arguments))
            pred.perform_mapping(context, accumulator)



import logging
import numbers
from copy import copy
from pathlib import Path
from typing import Any, IO, Iterable, Iterator, List, Mapping, Optional, Sequence, Union  # noqa
from .helper.typing import ClosableIterable
from .solver import DefaultSolver, Solver, SolverOptions
from .helper import CachingIterable
from .input import InputSpec, FactAccumulator
from .output import UndefinedNameError, OutputSpec
from .parser import parse_embedded_spec
from .registry import Registry, global_registry
from . import asp

__all__ = ['Program']

log = logging.getLogger(__name__)


class Program:
    '''Represents an answer set program.'''

    def __init__(self,
                 *,
                 filename: Optional[Union[str, Path]] = None,
                 code: Optional[str] = None,
                 use_global_registry: bool = True) -> None:
        '''Initialize an answer set program.

        For convenience, calls the appropriate `append...` method with the given `filename` and `code` keyword arguments.
        '''
        self.file_parts = []  # type: List[str]
        self.code_parts = []  # type: List[str]
        self._input_spec = None  # type: Optional[InputSpec]
        self._output_spec = None  # type: Optional[OutputSpec]
        self.solver = None  # type: Solver
        self.local_registry = copy(global_registry) if use_global_registry else Registry()  # type: Registry
        if filename is not None:
            self.append_file(filename)
        if code is not None:
            self.append_code(code)

    def __copy__(self) -> 'Program':
        other = Program()
        other.file_parts = copy(self.file_parts)
        other.code_parts = copy(self.code_parts)
        other._input_spec = copy(self._input_spec)
        other._output_spec = copy(self._output_spec)
        other.solver = copy(self.solver)
        other.local_registry = copy(self.local_registry)
        return other

    @property
    def local_registry(self) -> Registry:
        return self._local_registry

    @local_registry.setter
    def local_registry(self, value: Registry) -> None:
        self._local_registry = value
        self.register = self._local_registry.register
        self.register_dict = self._local_registry.register_dict
        self.import_from_module = self._local_registry.import_from_module

    @property
    def input_spec(self) -> InputSpec:
        '''The input specification, if one has been set or parsed from input, or an empty specification otherwise.'''
        if self._input_spec is not None:
            return self._input_spec
        else:
            return InputSpec.empty()

    @input_spec.setter
    def input_spec(self, value: InputSpec) -> None:
        self._input_spec = value

    @property
    def has_input_spec(self) -> bool:
        '''True iff an input specification has been set.'''
        return self._input_spec is not None

    @property
    def output_spec(self) -> OutputSpec:
        '''The output specification, if one has been set or parsed from output, or an empty specification otherwise.'''
        if self._output_spec is not None:
            return self._output_spec
        else:
            return OutputSpec.empty()

    @output_spec.setter
    def output_spec(self, value: OutputSpec) -> None:
        self._output_spec = value

    @property
    def has_output_spec(self) -> bool:
        '''True iff an output specification has been set.'''
        return self._output_spec is not None

    def parse_spec(self, code: str) -> None:
        i, o = parse_embedded_spec(code)
        if i is not None:
            if not self.has_input_spec:
                self.input_spec = i
            else:
                raise ValueError("Only one INPUT specification per program is allowed.")
        if o is not None:
            if not self.has_output_spec:
                self.output_spec = o
            else:
                raise ValueError("Only one OUTPUT specification per program is allowed.")

    def append_file(self, filename: Union[str, Path], *, parse_io_spec: bool = True, encoding: str = 'UTF-8') -> None:
        '''Append ASP code contained in the given file to the program.

        Note that unless the `parse_io_spec` argument is `False`, the file is opened immediately to extract any embedded I/O specifications.
        The file path is passed to the solver that will read the actual ASP code at the time of solving.
        '''
        filename = str(filename)  # also support pathlib.Path instances
        # TODO: If the encoding differs from what the solver expects, we should just read the file and append it to the code parts
        self.file_parts.append(filename)
        if parse_io_spec:
            with open(filename, 'rt', encoding=encoding) as file:
                self.parse_spec(file.read())

    def append_code(self, code: str, *, parse_io_spec: bool = True) -> None:
        '''Append the given ASP code to the program.

        Unless the `parse_io_spec` argument is `False`, any embedded I/O specifications are extracted from the given string.
        '''
        self.code_parts.append(code)
        if parse_io_spec:
            self.parse_spec(code)

    def solve(self,
              *input_arguments,
              solver: Optional[Solver] = None,
              options: Optional[SolverOptions] = None,
              cache: bool = True) -> 'Results':
        '''Solve the ASP program with the given input arguments and return a collection of answer sets.

        If deterministic cleanup of the solver subprocess is required, call close() on the returned object,
        or use the returned object as a context manager in a `with` statement.
        '''
        if solver is None:
            solver = self.solver
            if solver is None:
                solver = DefaultSolver()

        def write_asp_input(text_stream: IO[str]) -> None:
            '''Write all facts and rules that are needed in addition to the original ASP code to the given stream.'''
            # Map input data and pass it over the stream
            # Raises exception if the input arguments are not as expected (e.g., wrong count, an attribute does not exist, ...)
            self.input_spec.perform_mapping(input_arguments, StreamAccumulator(text_stream))
            # Additional rules required for output mapping
            for rule in self.output_spec.additional_rules():
                log.debug('Program: Adding helper rule %r', rule)
                text_stream.write(str(rule))
                text_stream.write('\n')
            # Pass code given as string over stdin
            for code in self.code_parts:
                text_stream.write(code)

        answer_sets = solver.run(
            write_input=write_asp_input,
            capture_predicates=self.output_spec.captured_predicates(),
            file_args=self.file_parts,
            options=options
        )
        return Results(answer_sets, self.output_spec, self.local_registry, cache)

    def solve_one(self,
                  *input_arguments,
                  solver: Optional[Solver] = None,
                  options: Optional[SolverOptions] = None) -> Optional['Result']:
        '''Solve the ASP program and return one of the computed answer sets, or None if no answer set exists. No special cleanup is necessary.'''
        options = SolverOptions() if options is None else copy(options)
        options.max_answer_sets = 1
        with self.solve(*input_arguments, solver=solver, options=options, cache=False) as results:
            try:
                return next(iter(results))
            except StopIteration:
                return None


class StreamAccumulator(FactAccumulator):
    def __init__(self, output_stream: IO[str]) -> None:
        if not output_stream.writable:
            raise ValueError('output_stream must be writable')
        self._stream = output_stream

    def arg_str(self, arg: Any) -> str:
        '''Convert the given argument to a string suitable to be passed to the ASP solver.'''
        if isinstance(arg, numbers.Integral):
            # Output integers without quotes (so we can use them for arithmetic in ASP)
            return str(arg)
        else:
            # Everything else is converted to a string and quoted unconditionally
            return asp.quote(arg)

    def add_fact(self, predicate: str, args: Sequence[Any]) -> None:
        '''Writes a fact to the output stream, in the usual ASP syntax: predicate(arg1, arg2, arg3).'''
        assert len(predicate) > 0
        self._stream.write(predicate)
        self._stream.write('(')
        for (idx, arg) in enumerate(args):
            if idx > 0:
                self._stream.write(',')
            self._stream.write(self.arg_str(arg))
        self._stream.write(').\n')
        if log.isEnabledFor(logging.DEBUG):  # type: ignore
            fact = predicate + '(' + ', '.join(self.arg_str(x) for x in args) + ')'
            log.debug('StreamAccumulator: Adding fact for predicate %r with args %r:\t=> %s', predicate, args, fact)


class Results(Iterable['Result']):
    '''The collection of results of a Solver invocation, corresponding to the set of all answer sets.'''
    # TODO: Describe implicit access to mapped objects through __getattr__ (e.g. .all_graph iterates over answer sets, returning the "graph" object for every answer set)
    # TODO: Should support async/await

    def __init__(self, answer_sets: ClosableIterable[asp.RawAnswerSet], output_spec: OutputSpec, registry: Registry, cache: bool) -> None:
        self.output_spec = output_spec
        self.registry = registry
        self.answer_sets = answer_sets
        self.results = (
            Result(answer_set, self.output_spec, self.registry) for answer_set in self.answer_sets
        )  # type: Iterable[Result]
        if cache:
            self.results = CachingIterable(self.results)

    def __iter__(self) -> Iterator['Result']:
        # Make sure we can only create one results iterator if we aren't caching
        assert self.results is not None, 'Pass cache=True if you need to iterate over results multiple times.'
        rs = self.results
        if type(self.results) != CachingIterable:
            self.results = None
        yield from rs

    def __bool__(self) -> bool:
        # TODO: Handle case when cache=False
        # if not self.cache:
        #     raise NotImplementedError('Because cache=False, only (one) iteration is supported.')
        try:
            next(iter(self))
            return True
        except StopIteration:
            return False

    def __getattr__(self, name: str) -> Any:
        prefix = 'each_'
        if name.startswith(prefix):
            return ResultsAttributeIterator(self, name[len(prefix):])
        else:
            raise AttributeError("No attribute with name {0!r}. Prefix an output variable name with '{1!s}' when iterating over its values for all answer sets.".format(name, prefix))

    def close(self) -> None:
        self.answer_sets.close()

    def __enter__(self) -> 'Results':
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
        return False


class ResultsAttributeIterator(Iterable[Any]):
    '''Helper class to iterate over the values of the output variable with the given name for all answer sets.'''
    # This class is required to support explicit cleanup with 'all_' shortcuts when accessing the output data.

    def __init__(self, results: Results, name: str) -> None:
        self.results = results
        self.name = name

    def __iter__(self) -> Iterator[Any]:
        return iter(getattr(r, self.name) for r in self.results)

    def close(self) -> None:
        self.results.close()

    def __enter__(self) -> 'ResultsAttributeIterator':
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
        return False


class Result:
    '''Represents a single answer set.'''

    def __init__(self, answer_set: asp.RawAnswerSet, output_spec: OutputSpec, registry: Registry) -> None:
        self.answer_set = answer_set
        self._r = output_spec.prepare_mapping(answer_set, registry)

    def get(self, name: str) -> Any:
        return self._r.get_object(name)

    def __getattr__(self, name: str) -> Any:
        try:
            return self.get(name)
        except UndefinedNameError as e:
            raise AttributeError(e)



from typing import Generic, Iterable, Iterator, TypeVar

__all__ = ['CachingIterable']


T = TypeVar('T', covariant=True)


class CachingIterable(Generic[T], Iterable[T]):
    '''Caches the contents of an iterable.

    Similar to calling list() on an iterable.
    However, where list() would finish the whole iteration before returning,
    the CachingIterable only advances the underlying iterator when the element is actually requested.

    Supports multiple iterators,
    but calls to next() on iterators constructed from instances of this class
    must be synchronized if they are to be used from multiple threads.
    '''

    def __init__(self, base_iterable: Iterable[T]) -> None:
        self.base_iterator = iter(base_iterable)
        self.cache = []  # type: List[T]
        self.done = False

    def __iter__(self) -> Iterator[T]:
        pos = 0
        while True:
            if pos < len(self.cache):
                yield self.cache[pos]
                pos += 1
            elif self.done:
                break
            else:
                self._generate_next()

    def _generate_next(self) -> None:
        if not self.done:
            try:
                value = next(self.base_iterator)
                self.cache.append(value)
            except StopIteration:
                self.done = True
                del self.base_iterator  # release underlying object



from .abc import Solver, SolverOptions
from .dlvhex2 import Dlvhex2Solver

__all__ = [
    'DefaultSolver',
    'Dlvhex2Solver',
    'Solver',
    'SolverOptions',
]


def DefaultSolver() -> Solver:
    """Returns an instance of the default solver."""
    # TODO: Once we support multiple solvers, we should set the default solver depending on what's installed
    return Dlvhex2Solver()



import logging
from .errors import CircularReferenceError, InvalidIndicesError, RedefinedNameError, SolverError, UndefinedNameError
from .input import InputSpec
from .output import OutputSpec
from .program import Program
from .registry import register, register_dict, import_from_module
from .solver import Solver, SolverOptions

__all__ = [
    'CircularReferenceError',
    'InvalidIndicesError',
    'RedefinedNameError',
    'SolverError',
    'UndefinedNameError',
    #
    'InputSpec',
    #
    'OutputSpec',
    #
    'Program',
    #
    'register',
    'register_dict',
    'import_from_module',
    #
    'Solver',
    'SolverOptions',
]

# Set up logging. By default, do not output any log messages from library code.
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())



from setuptools import setup, find_packages

setup(
    name='aspio',
    version='0.0.1',
    description='ASP interface to object-oriented programs',
    long_description='This library simplifies integration of Answer Set Programming into Python applications.',
    url='https://github.com/hexhex/py-aspio',

    packages=find_packages(exclude=['*.tests', '*.tests.*', 'tests.*', 'tests']),

    install_requires=[
        'pyparsing >= 2.1.4',
    ],

    author='Jakob Rath',
    author_email='jakob.rath@student.tuwien.ac.at',
    license='MIT',

    test_suite='aspio.tests',

    # A list of classifiers can be found at
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
    ],
)



from collections import namedtuple
from typing import Any, Iterable, Mapping, Tuple, Union


class Variable(namedtuple('Variable', ['name'])):
    '''Represents a variable in ASP syntax.'''
    __slots__ = ()

    def __str__(self):
        return self.name


class AnonymousVariable():
    '''Represents an anonymous variable in ASP syntax.'''
    __slots__ = ()

    def __str__(self):
        return '_'


def quote(arg: Any) -> str:
    '''Enclose the given argument in double quotes, escaping any contained quotes and backslashes with a backslash.'''
    return '"' + str(arg).replace('\\', '\\\\').replace('\"', '\\\"') + '"'


# dlvhex2 distinguishes between unquoted and quoted strings,
# i.e. abc and "abc" are different constant symbols,
# or in other words, the program
#       p(abc). -p("abc").
# has an answer set.
# We must preserve this distinction in the query.
# Unquoted constants are represented by `str`, and quoted constants by `QuotedConstant`.
class QuotedConstant(namedtuple('QuotedConstant', ['value'])):
    '''Represents a quoted string constant in ASP syntax.'''
    __slots__ = ()

    def __str__(self):
        return quote(self.value)

Term = Union[int, str, QuotedConstant, Variable]
TermTuple = Tuple[Term, ...]


class Literal(namedtuple('Literal', ['predicate', 'arguments', 'defaultNegated'])):
    '''A literal, not necessarily ground, possibly default-negated'''
    __slots__ = ()

    def __str__(self):
        return ('not ' if self.defaultNegated else '') + self.predicate + '(' + ','.join(str(t) for t in self.arguments) + ')'

    def variables(self) -> Iterable[Variable]:
        for arg in self.arguments:
            if isinstance(arg, Variable):
                yield arg

Literal.predicate.__doc__ = 'predicate name (prefixed with a `-` for strongly negated literals)'
Literal.arguments.__doc__ = 'a tuple containing the argument terms, cf. type `TermTuple`'
Literal.defaultNegated.__doc__ = 'a bool determining whether the literal is default-negated'


class Query:
    '''A set of (possibly non-ground) literals, representing a query.'''

    def __init__(self, literals: Tuple[Literal, ...]) -> None:
        self.literals = literals

    def __str__(self) -> str:
        return ','.join(str(t) for t in self.literals)

    def variables(self) -> Iterable[Variable]:
        for lit in self.literals:
            yield from lit.variables()


Rule = str  # TODO: Should be a more sophisticated type to support passing the rule to the solver directly (when it is used via a shared library)

# "raw" answer set, i.e. the constants are just strings
RawAnswerSet = Mapping[str, Iterable[Tuple[str, ...]]]
